// Package forge implements sort of a stack-based interpreter.
package forge

import (
	"git.sr.ht/~cco/go-scopes/common/ptr"
	"git.sr.ht/~cco/go-scopes/common/stack"
	"git.sr.ht/~cco/go-scopes/common/voc"
)

type fitem interface{}

type fptr = ptr.Ptr[fitem]
type fstack = stack.Stack[fitem]
type fvoc = voc.Vocabulary[fitem]

type Callable func(*forgeEnv, XT)

type forgeEnv struct {
	ds, rs     fstack
	cp, ip, dp fptr
	voc        *fvoc
}

type FE = *forgeEnv

func NewFE() *forgeEnv {
	return newFE(newVoc(nil))
}

func (f *forgeEnv) ChildFE() *forgeEnv {
	return newFE(newVoc(f.voc))
}

func newFE(voc *fvoc) *forgeEnv {
	return &forgeEnv{
		ds:  stack.NewStack[fitem](),
		rs:  stack.NewStack[fitem](),
		ip:  ptr.NewSlice[fitem](),
		voc: voc,
	}
}

var newVoc = voc.NewVoc[fitem]

var newScalar = ptr.NewScalar[fitem]
var newPtr = ptr.NewSlice[fitem]

// XT (execution token)

type XT interface {
	Name() string
	Fct() Callable
	Code() fptr
	Data() fptr
}

type name string

func (n name) Name() string {
	return string(n)
}

// gofunc: XT consisting only of a function written in Go

type gofunc struct {
	name
	fct Callable
}

func GoFunc(n string, fct Callable) XT {
	return &gofunc{name(n), fct}
}

func (it *gofunc) Fct() Callable {
	return it.fct
}

func (it *gofunc) Code() fptr {
	return nil
}

func (it *gofunc) Data() fptr {
	return nil
}

// xitem: executable item with all parts

type xitem struct {
	name
	fct  Callable
	body fptr
	code fptr
	data fptr
}

func (it *xitem) Fct() Callable {
	return it.fct
}

func (it *xitem) Code() fptr {
	//return it.code
	return it.body
}

func (it *xitem) Data() fptr {
	//return it.data
	return it.body
}

func Register(voc *fvoc, n string, fct Callable) *xitem {
	it := xitem{
		name: name(n),
		fct:  fct,
	}
	voc.Register(n, &it)
	return &it
}

// forgeEnv methods

func (f *forgeEnv) Code(items ...fitem) fptr {
	code := newPtr(items...)
	// ... pre-process (compile) code ...
	return code
}

func (f *forgeEnv) Call(code fptr) {
	f.rs.Push(f.ip)
	f.ip = code.Clone()
	for f.ip.Next() != nil {
		xt := f.ip.Value().(XT)
		xt.Fct()(f, xt)
	}
	f.ip = f.rs.Pop().(fptr)
}

func (f *forgeEnv) Exec(items ...fitem) {
	f.Call(f.Code(items...))
}

func (f *forgeEnv) Def(name string, items ...fitem) XT {
	xt := Register(f.voc, name, doDef)
	xt.body = newPtr(f.Code(items...)).Next()
	return xt
}

func (f *forgeEnv) Const(name string, value fitem) XT {
	xt := Register(f.voc, name, doData)
	xt.body = newScalar()
	xt.body.Append(value)
	return xt
}

func (f *forgeEnv) Literal() {
	f.Push(f.ip.Next().Value())
}

func (f *forgeEnv) Voc() *fvoc {
	return f.voc
}

func (f *forgeEnv) Push(it fitem) {
	f.ds.Push(it)
}

func (f *forgeEnv) Pop() fitem {
	return f.ds.Pop()
}

func (f *forgeEnv) Peek(d int) fitem {
	return f.ds.Peek(d)
}

// basic functions for executable items

func doDef(f *forgeEnv, xt XT) {
	f.Call(xt.Code().Value().(fptr))
}

func doData(f *forgeEnv, xt XT) {
	f.Push(xt.Data().Value())
}
